#!/usr/bin/perl -w

#  irspider v1.1.0 (Internet Relay [chat] Spider) 
#  Copyright (C) 2000, Adam J. Foxson  
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
#  This program is developed and maintained by Adam J. Foxson.
#
#  If you have any questions, suggestions, problems,
#  inquiries or otherwise, please feel free to to
#  contact me at: afoxson@guild.net
#
#  irspider webpages:
#
#  https://sourceforge.net/project/?group_id=3410
#  http://irspider.sourceforge.net

# $Revision: 1.3 $
# $Id: irspider.pl,v 1.3 2000/05/14 11:45:09 afoxson Exp $

use strict;
use POE::Kernel;
use POE::Session;
use POE::Component::IRC;
use vars qw(@servers $server %config $kernel @adminhosts);

# Shortcut variables
my $start           = '-\*([a-zA-Z\-0-9]+)\*-\s';
my $t               = '\*\*';
my $s               = '\s\s';
my $d               = '(\d+)';

# Regular expression variables
my $match_stats0    =
	qr(${start}\[\*\sStat\s\*]\s$d\sfiles?\.\s\s$d\sof\s$d\sslots?\sopen);
my $match_stats1    =
	qr($start$t\s$d\spacks?\s$t$s$d\sof\s$d\sslots\sopen,\sRecord:\s(.*));
my $match_stats2    =
	qr($start$t\s$d\spacks?\s$t$s$d\sof\s$d\sslots\sopen,\s(Max|Min):\s(.+),\sRecord:\s(.*));
my $match_stats3    =
	qr($start$t\s$d\spacks?\s$t$s$d\sof\s$d\sslots\sopen,\sQueue:\s(\d+)/(\d+));
my $match_bandwidth1 =
	qr($start$t\sBandwidth\sUsage\s$t\sCurrent:\s(.*),\sCap:\s(.*),\sRecord:\s(.*));
my $match_bandwidth2 =
	qr($start$t\sBandwidth\sUsage\s$t\sCurrent:\s(.*),\sRecord:\s(.*));
my $match_pack      =
	qr($start\#$d\s+${d}x\s+\[\s*(.+?)\]\s+(.*));
my $match_note      =
	qr($start\s\^-\s+(.*));
my $match_totals    =
	qr(${start}Total\sOffered:\s(.*)${s}Total\s(?:Transferred|Snagged):\s(.*));
my $match_bot_nick  =
	qr([Dd][Cc][Cc]\d*);

# global lexical variables
my %bot             = ();
my @requestors      = ();
my %flooders        = ();
my $pack            = 0;
my $start_time      = time;
my $total_requests  = 0;
my $version         = '1.1.0';

&process_configfile(1);

# tuck these variables away for later timer related use
my $respider_save = $config{respider};
my $plist_save = $config{plist};

# setup logging..
if ($config{log})
{
	open(LOGFILE, ">>./irspider.log") or
		die "Error appending logfile: $!";
	select LOGFILE;
	$|++;
	open(STDERR, ">>./irspider.log") or
		die "Error appending logfile: $!";
	select STDERR;
	$|++;
	select STDOUT;
}

# write out a PID file
open(PIDFILE, ">irspider.pid") or
	die "Error creating PID file: $!";
print PIDFILE $$;
close(PIDFILE);

# pick a random server to connect to
$server=&pickserver();
sub pickserver
{
	return $servers[ rand scalar @servers ];
}

sub _start
{
	my ($kernel, $session) = @_[KERNEL, SESSION];
	# $session->option( trace => 1 );

	$kernel->alias_set('irspider');
	$kernel->post( 'to_irspider', 'register',
		qw(001 kick disconnected error msg notice 353 join nick part quit));
	$kernel->post( 'to_irspider', 'connect',
	{
		Debug    => 0,
		Nick     => $config{nick},
		Server   => $server,
		Port     => $config{port},
		Username => $config{nick},
		Ircname  => "/msg $config{nick} help",
	});
}

sub irc_001
{
	my ($kernel) = $_[KERNEL];

	mylog("connect: successfull connect to $server.\n");

	$kernel->post( 'to_irspider', 'mode', $config{nick}, '+i' );
	$kernel->post( 'to_irspider', 'join', $config{channel} );
	$kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
		"** $config{nick} (irspider v$version) **  IRC Search Engine Smartbot");
	$kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
		"** To request a search: \"/msg $config{nick} search <whatever>\"");
	$kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
		"** To request help: \"/msg $config{nick} help\"");
	#$kernel->post( 'to_irspider', 'away', 'I am away away away');
}

sub irc_353
{
	my ($kernel) = $_[KERNEL];
	my ($names)  = $_[ARG1];
	my @names    = split(/\s+/, $names);

	for (@names)
	{
		if ($_ =~ /$match_bot_nick$/i)
		{
			$_ = substr($_, 1) if (substr($_, 0, 1) eq ':');
			if ((substr($_, 0, 1) eq '@') or
				(substr($_, 0, 1) eq '+'))
			{
				$_ = substr($_, 1);
			}
			if ($_ !~ /$config{nick}/i)
			{
				mylog("spider: on connect: $_\n");
				$kernel->post( 'to_irspider', 'privmsg', $_, 'xdcc list' );
			}
		}
	}
}


sub irc_msg
{
	local ($kernel) = $_[KERNEL];
	my ($sender, $to, $text) = @_[ARG0 .. $#_];
	my $hostmask = $sender;

	# strip hostmask info, we want the nick only
	$sender =~ s/(.+)!.+/$1/;

	if ($flooders{$sender} and $flooders{$sender} > 5)
	{
		if ($flooders{$sender} == 6)
		{
			$kernel->post('to_irspider', 'notice', $sender,
				'You seem to be abusing me. All further requests from you');
			$kernel->post('to_irspider', 'notice', $sender,
				'will now be dropped for this session.');
			mylog("abuse control: enforced on: $sender\n");
		}
		$flooders{$sender} = $flooders{$sender} + 1;
		return;
	}

	mylog("message: received from $sender: $text\n");
	$total_requests++;

	# check to see if this person seems to be flooding us, and if so
	# return an error
	my $iteration = 0;
	for (0 .. $#requestors)
	{
		next if $_ % 2 != 0;
		if ($sender eq $requestors[$_])
		{
			$iteration = $_ + 1;
			if ((time - $requestors[$iteration]) <= $config{antiflood})
			{
				$flooders{$sender} = 0 if not $flooders{$sender};
				$flooders{$sender} = $flooders{$sender} + 1;
				$kernel->post('to_irspider', 'notice', $sender,
					'Error: You must wait a few seconds between each request.');
				return;
			}
		}
	}

	# if the user is already in the requestor list, let's simply update
	# their current entry..
	my $success = 0;
	for (0 .. $#requestors)
	{
		next if $_ % 2 != 0;
		if ($sender eq $requestors[$_])
		{
			$iteration = $_ + 1;
			$requestors[$iteration] = time;
			$success++;	
		}
	}

	# if the user is not in the requestor list, let's remove the last
	# user/time entry, and add the current person to the front
	if (not $success)
	{
		pop @requestors;
		pop @requestors;
		unshift @requestors, time;
		unshift @requestors, $sender;
	}

	if ($text eq 'ping')
	{
		$kernel->post('to_irspider', 'notice', $sender, 'PONG');
		return;
	}
	elsif ($text eq 'bots')
	{
		&show_bots($sender, 1);
		return;
	}
	elsif ($text eq 'stats')
	{
		&show_stats($sender);
		return;
	}
	elsif ($text eq 'shutdown')
	{
		return if not &verify_admin($hostmask, 1);
		mylog("shutdown: per request of $sender.\n");
		$kernel->post( 'to_irspider', 'quit', 'brb!');
		$kernel->alias_remove( 'irspider' );
		unlink('irspider.pid');
		sleep 2;
		die;
	}
	elsif ($text eq 'help')
	{
		&help($sender, $hostmask);
		return;
	}
	elsif ($text eq 'rehash')
	{
		return if not &verify_admin($hostmask, 1);
		&process_configfile(2);
		$kernel->post('to_irspider', 'notice', $sender,
			'Rehashed configuration file.');
		return;
	}

	my ($cmd, $args) = split(/\s+/, $text);

	if ($cmd eq 'search')
	{
		return if not &enough_args($sender, $args);
		&master_find($args, $sender);
	}
	elsif ($cmd eq 'add')
	{
		return if not &enough_args($sender, $args);
		return if not &verify_admin($hostmask, 1);
		$kernel->post( 'to_irspider', 'privmsg', $args, 'xdcc list' );
		$kernel->post('to_irspider', 'notice', $sender, "$args spidered");
	}
	elsif ($cmd eq 'del')
	{
		return if not &enough_args($sender, $args);
		return if not &verify_admin($hostmask, 1);
		if ($bot{$args})
		{
			delete $bot{$args};
			$kernel->post('to_irspider', 'notice', $sender, "$args removed");
		}
		else
		{
			$kernel->post('to_irspider', 'notice', $sender,
				'I am not tracking that bot!');
		}
	}
	elsif ($cmd eq 'masslist')
	{
		return if not &enough_args($sender, $args);
		&masslist($sender, $args);
	}
	elsif ($cmd eq 'info')
	{
		return if not &enough_args($sender, $args);
		&report_info($args, $sender);
	}
	elsif ($cmd eq 'xdcc')
	{
		if ($args eq 'list')
		{
			$kernel->post( 'to_irspider', 'notice', $sender,
				"** $config{nick} (irspider v$version) **  IRC Search Engine Smartbot");
			$kernel->post( 'to_irspider', 'notice', $sender,
				"** To request a search: \"/msg $config{nick} search <whatever>\"");
			$kernel->post( 'to_irspider', 'notice', $sender,
				"** To request help: \"/msg $config{nick} help\"");
		}
	}
	else
	{
		$kernel->post( 'to_irspider', 'notice', $sender,
			'Invalid command. Valid commands are:');
		$kernel->post( 'to_irspider', 'notice', $sender,
			'ping, search, info, masslist, bots, stats, help, add, del, shutdown, rehash');
		$kernel->post( 'to_irspider', 'notice', $sender,
			'Use the \'help\' command for detailed info. on each command');
	}

	sub enough_args
	{
		my $sender = shift;
		my $argument = shift;

		if (not $argument)
		{
			$kernel->post('to_irspider', 'notice', $sender,
				"Error: This command requires an argument. Try again");
			return 0;
		}
		else
		{
			return 1;
		}
	}


	sub help
	{
		my $sender   = shift;
		my $hostmask = shift;

		$kernel->post( 'to_irspider', 'notice', $sender,
			'Commands are:');
		$kernel->post( 'to_irspider', 'notice', $sender,
			' ');
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} search <whatever> - to search for something");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} info <botnick> - get detailed info on a bot");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} masslist web - to get a web masslist");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} masslist dcc - to get a dcc masslist (not yet available)");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} bots - listing of bots currently tracked");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} stats - list internal bot statistics");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} help - list all commands, and usage");

		return if not &verify_admin($hostmask, 2);

		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} add <bot> - spider a bot manually (admin)");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} del <bot> - despider a bot manually (admin)");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} shutdown <message> - shutdown the bot (admin)");
		$kernel->post( 'to_irspider', 'notice', $sender,
			"/msg $config{nick} rehash - rehashes the configfile (admin)");
	}
}

sub irc_notice
{
	my ($kernel) = $_[KERNEL];
	my ($sender, $to, $text) = @_[ARG0 .. $#_];
	my $nick = '';

	# Strip out the hostmask, so we are left only with a nick type that
	# we want to work with
	$sender =~ s/(.+)!.+/-\*$1\*-/;
	$nick = $sender;
	chop $nick; chop $nick;
	$nick = substr $nick, 2;

	# If the person who is sending this notice does not appear to be a
	# bot, don't bother trying to parse it's notices to us.
	return if $nick !~ /$match_bot_nick$/i;    

	# Let's try to see if we can parse this bot's notice to us
	&master_match("$sender $text");
}


sub _stop
{
	my ($kernel) = $_[KERNEL];

	mylog("disconnect: control session stopped.\n");
	$kernel->post( 'to_irspider', 'quit', 'brb!');
	$kernel->alias_remove( 'irspider' );
	unlink('irspider.pid');
}


sub irc_disconnected
{
	my ($server) = $_[ARG0];
	mylog("disconnect: lost connection to server $server.\n");
}


sub irc_error
{
	my $err = $_[ARG0];
	mylog("error: server error occurred! $err\n");
}


sub irc_kick
{
	my ($kicker, $channel, $kickee, $reason) = @_[ARG0 .. ARG4];

	if ($bot{$kickee})
	{
		delete $bot{$kickee};
		mylog("kick: deleted bot $kickee on kick by $kicker\n");
	}
}

sub irc_join
{
	local ($kernel) = $_[KERNEL];
	my ($nick, $channel) = @_[ARG0 .. ARG4];
	$nick =~ s/(.+)!.+/$1/;

	if ($nick =~ /$match_bot_nick$/i)
	{
		if ($nick !~ /$config{nick}/i)
		{
			$kernel->post('to_irspider', 'privmsg', $nick, 'xdcc list');
			mylog("spider: on join: $nick\n");
		}
	}
}

sub irc_nick
{
	local ($kernel) = $_[KERNEL];
	my ($old_nick, $new_nick) = @_[ARG0 .. ARG4];
	$old_nick =~ s/(.+)!.+/$1/;

	if ($new_nick =~ /$match_bot_nick$/i)
	{
		delete $bot{$old_nick} if $bot{$old_nick};
		$kernel->post('to_irspider', 'privmsg', $new_nick, 'xdcc list');
		mylog("nick: from $old_nick to $new_nick\n");
	}
}

sub irc_part
{
	my ($nick, $channel) = @_[ARG0 .. ARG4];
	$nick =~ s/(.+)!.+/$1/;

	if ($bot{$nick})
	{
		delete $bot{$nick};
		mylog("part: deleted bot $nick on part\n");
	}
}

sub irc_quit
{
	my ($nick, $reason) = @_[ARG0 .. ARG4];
	$nick =~ s/(.+)!.+/$1/;

	if ($bot{$nick})
	{
		delete $bot{$nick};
		mylog("quit: deleted bot $nick on quit\n");
	}
}

sub master_match
{
	my $text = shift;
	$text =~ tr[\x00-\x1f][]d;

	if ($text =~ /$match_pack/)
	{
		$bot{$1}{Pack}[$2]{Description}  = $5;
		$bot{$1}{Pack}[$2]{Downloads}    = $3;
		$bot{$1}{Pack}[$2]{Size}         = $4;
		$pack = $2;
	}
	elsif ($text =~ /$match_stats0/)
	{
		$bot{$1}{TotalPacks}             = $2;
		$bot{$1}{SlotsCurrentlyAvail}    = $3;
		$bot{$1}{SlotsTotalAvailable}    = $4;
	}
	elsif ($text =~ /$match_stats1/)
	{
		$bot{$1}{TotalPacks}             = $2;
		$bot{$1}{SlotsCurrentlyAvail}    = $3;
		$bot{$1}{SlotsTotalAvailable}    = $4;
		$bot{$1}{BandwidthOverallRecord} = $5;
	}
	elsif ($text =~ /$match_stats2/)
	{
		$bot{$1}{TotalPacks}             = $2;
		$bot{$1}{SlotsCurrentlyAvail}    = $3;
		$bot{$1}{SlotsTotalAvailable}    = $4;
		$bot{$1}{BandwidthMinMax}        = $5;
		$bot{$1}{BandwidthMinMaxActual}  = $6;
		$bot{$1}{BandwidthOverallRecord} = $7;
	}
	elsif ($text =~ /$match_stats3/)
	{
		$bot{$1}{TotalPacks}             = $2;
		$bot{$1}{SlotsCurrentlyAvail}    = $3;
		$bot{$1}{SlotsTotalAvailable}    = $4;
		$bot{$1}{QueueCurrentlyUsed}     = $5;
		$bot{$1}{QueueTotalAvailable}    = $6;
	}
	elsif ($text =~ /$match_totals/)
	{
		$bot{$1}{TotalOffered}           = $2;
		$bot{$1}{TotalTransferred}       = $3;
	}
	elsif ($text =~ /$match_note/)
	{
		$bot{$1}{Pack}[$pack]{Note}      = $2;
		$pack = 0;
	}
	elsif ($text =~ /$match_bandwidth1/)
	{
		$bot{$1}{BandwidthCurrantUsage}  = $2;
		$bot{$1}{BandwidthCurrentCap}    = $3;
		$bot{$1}{BandwidthCurrentRecord} = $4;
	}
	elsif ($text =~ /$match_bandwidth2/)
	{
		$bot{$1}{BandwidthCurrantUsage}  = $2;
		$bot{$1}{BandwidthCurrentRecord} = $3;
	}
	#else
	#{
	#	mylog("notice: unrecognized bot notice: $text\n");
	#}
}

sub report_info
{
	my $input    = shift;
	my $sender   = shift;
	my $success  = 0;

	if (not $input or length($input) < 1)
	{
		$kernel->post('to_irspider', 'notice', $sender,
			'Error: You must specify a bot to get info on.');
		return;
	}

	for my $b (sort (keys %bot))
	{
		if ($b =~ /^$input$/i)
		{
			$success++;
			$kernel->post('to_irspider', 'notice', $sender,
				"Detailed information on the bot: $b");
			$kernel->post('to_irspider', 'notice', $sender,
				' ');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Total Packs         : %s\n",
					($bot{$b}{TotalPacks})?$bot{$b}{TotalPacks}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Slots Avail         : %s\n",
					($bot{$b}{SlotsCurrentlyAvail})?$bot{$b}{SlotsCurrentlyAvail}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Slots Total         : %s\n",
					($bot{$b}{SlotsTotalAvailable})?$bot{$b}{SlotsTotalAvailable}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Queue Used          : %s\n",
					($bot{$b}{QueueCurrentlyUsed})?$bot{$b}{QueueCurrentlyUsed}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Queue Avail         : %s\n",
					($bot{$b}{QueueTotalAvailable})?$bot{$b}{QueueTotalAvailable}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Band %s            : %s\n",
					($bot{$b}{BandwidthMinMax})?$bot{$b}{BandwidthMinMax}:'n/a',
					($bot{$b}{BandwidthMinMaxActual})?$bot{$b}{BandwidthMinMaxActual}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Fastest Ind Trnsfr  : %s\n",
					($bot{$b}{BandwidthOverallRecord})?$bot{$b}{BandwidthOverallRecord}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Band Current Usage  : %s\n",
					($bot{$b}{BandwidthCurrantUsage})?$bot{$b}{BandwidthCurrantUsage}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Band Current Cap    : %s\n",
					($bot{$b}{BandwidthCurrentCap})?$bot{$b}{BandwidthCurrentCap}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Band Current Rec    : %s\n",
					($bot{$b}{BandwidthCurrentRecord})?$bot{$b}{BandwidthCurrentRecord}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Total Offered       : %s\n",
					($bot{$b}{TotalOffered})?$bot{$b}{TotalOffered}:'n/a');
			$kernel->post('to_irspider', 'notice', $sender,
				sprintf "Total Transferred   : %s\n",
					($bot{$b}{TotalTransferred})?$bot{$b}{TotalTransferred}:'n/a');
		}
	}
	if (not $success)
	{
		$kernel->post('to_irspider', 'notice', $sender,
			'Sorry, no matches were found.');
		mylog("info: reported no match to $sender.\n");
		return;
	}
	mylog("info: reported match to $sender.\n");
}

sub masslist
{
	# FIXME: File locking would probably be a good thing to have here.
	my $sender        = shift;
	my $type          = shift;
	my $file          = $config{masslist_location};
	my $count         = 0;
	my $time          = thetime(); 
	my $prelen      = 0;
	my $description = '';

	if ($type ne 'dcc' and $type ne 'web')
	{
		$kernel->post( 'to_irspider', 'notice', $sender,
			'That is an invalid masslist type. Use \'dcc\' or \'web\'.');
		$kernel->post( 'to_irspider', 'notice', $sender,
			'Please note that ONLY web masslists are currently supported.');
		return;
	}

	if ($type eq 'dcc')
	{
		$kernel->post( 'to_irspider', 'notice', $sender,
			'DCC masslists are not currently supported. Get a web masslist.');
		return;
	}

	open(DCCFILE, ">$file"), or die
		"Strange, cannot create the masslist file: $!";
	
	if ($type eq 'dcc')
	{
		print DCCFILE
			"--[ $config{channel} dynamically generated masslist ]--\r\n";
		print DCCFILE
			"--[ Generated at ${\(thetime())} by $config{nick} ]--\r\n";
		print DCCFILE
			"--[ Best viewed with a small-point fixed width font ]--\r\n\r\n";
    	print DCCFILE "File                                              Bot        Pack Dwnlds Size\r\n\r\n";
	}
	else
	{
	print DCCFILE
		qq(
			<HTML>
				<HEAD>
					<TITLE>
						Masslist for $config{channel}, generated at: $time
					</TITLE>
				</HEAD>
				<BODY BGCOLOR="white">
					<BR>
					<H3>
						Masslist for $config{channel}, generated at: $time
					</H3>
					<BR>
					<TABLE BORDER="1">
						<TR>
							<TD>
								File Description
							</TD>
							<TD>
								Bot
							</TD>
							<TD>
								Pack
							</TD>
							<TD>
								Downloads
							</TD>
							<TD>
								Size
							</TD>
						</TR>
						<TR><TD></TD><TD></TD><TD></TD><TD></TD><TD></TD></TR>
		);
	}

	my @masslist = ();
	for my $c (sort (keys %bot))
	{
		for my $i (@{$bot{$c}{Pack}})
		{
			if ($i->{Description})
			{
				$count++;
				if ($type eq 'dcc')
				{
					$prelen = length($i->{Description});
					$description = substr($i->{Description}, 0, 45);
					$description = $description . '...' if $prelen > 45;
					printf DCCFILE
						("%-49s %-9s  %-3s  %-6s %-6s\r\n",
							$i->{Description},
							$c,
							$count,
							$i->{Downloads},
							$i->{Size}
						);
				}
				else
				{
					#print DCCFILE "<TR><TD>$i->{Description}</TD>\n";
					#print DCCFILE "<TD>$c</TD>\n";
					#print DCCFILE "<TD>$count</TD>\n";
					#print DCCFILE "<TD>$i->{Downloads}</TD>\n";
					#print DCCFILE "<TD>$i->{Size}</TD><TR>";
					# you do not see this, you do not see this
					push(@masslist,
						"$i->{Description}|||||" .
						"$c|||||" .
						"$count|||||" .
						"$i->{Downloads}|||||" .
						"$i->{Size}");
				}
			}
		}
		$count = 0;
	}

	for (sort @masslist)
	{
		my ($description, $bot, $pack, $downloads, $size) =
			split(/\|\|\|\|\|/, $_);	
		print DCCFILE "<TR><TD>$description</TD>\n";
		print DCCFILE "<TD>$bot</TD>\n";
		print DCCFILE "<TD>$pack</TD>\n";
		print DCCFILE "<TD>$downloads</TD>\n";
		print DCCFILE "<TD>$size</TD><TR>";
	}

	if ($type ne 'dcc')
	{
		print DCCFILE
			qq(
						</TABLE>
					</BODY>
				</HTML>
			);
	}

	close(DCCFILE);

	if (-e $file)
	{
		if ($type eq 'dcc')
		{
			$kernel->post( 'to_irspider', 'notice', $sender,
				'Sending you the masslist via DCC now.');
			$kernel->post( 'to_irspider', 'dcc', $sender, 'send', $file);
		}
		else
		{
			$kernel->post( 'to_irspider', 'notice', $sender,
				"Masslist refreshed, visit: $config{masslist_response}");
			if ($config{masslist_username} and $config{masslist_password})
			{
				$kernel->post( 'to_irspider', 'notice', $sender,
				"When prompted the username is '$config{masslist_username}' " .
				"with password '$config{masslist_password}'.");
			}
		}
		mylog("masslist: sent successfully to $sender.\n");
	}
	else
	{
		$kernel->post( 'to_irspider', 'notice', $sender,
		'Strange, the masslist was not available. Try again shortly.');
		mylog("masslist: unable to send $sender a masslist. strange.\n");
	}
}

sub show_bots
{
	my $sender = shift;
	my $type   = shift;
	my $buffer;

	for my $b (sort (keys %bot))
	{
		if ($type == 1)
		{
			$buffer .= "$b ";	
		}
		elsif ($type == 2)
		{
			$buffer++;
		}
		else
		{
			die "Invalid show_bots type\n";
		}
	}
	
	if ($type == 1)
	{
		if ($buffer and length($buffer) > 1)
		{
			$kernel->post('to_irspider', 'notice', $sender,
				'Bots currently being tracked are: ');
			$kernel->post('to_irspider', 'notice', $sender, ' ');
			$kernel->post('to_irspider', 'notice', $sender, $buffer);
		}
		else
		{
			$kernel->post('to_irspider', 'notice', $sender,
				'No bots are currently being tracked');
		}
	}
	elsif ($type == 2)
	{
		return $buffer;
	}
}

sub show_packs
{
	my $count = 0;

	for my $b (sort (keys %bot))
	{
		for my $i (@{$bot{$b}{Pack}})
		{
			$count++ if $i->{Description};
		}
	}

	return $count;
}

sub show_stats
{
	my $sender = shift;
	my $since  = localtime($start_time);

	$kernel->post('to_irspider', 'notice', $sender,
		'-- Configuration file parameters --');
	$kernel->post('to_irspider', 'notice', $sender,
		sprintf("Logging: %s, Nick: %s, Channel: %s",
			($config{log})?'yes':'no',
			$config{nick},
			$config{channel}));
	$kernel->post('to_irspider', 'notice', $sender,
		"Plist: $plist_save (next: $config{plist} min), Respider: $respider_save (next: $config{respider} min)");
	$kernel->post('to_irspider', 'notice', $sender,
		"Maxreturns: $config{maxreturns}, Minlength: $config{minlength}, Antiflood: $config{antiflood}, IRC Port: $config{port}");
	$kernel->post('to_irspider', 'notice', $sender,
		'-- Statistics for this session --');
	$kernel->post('to_irspider', 'notice', $sender,
		"Total of $total_requests requests, since I started up at $since");
	$kernel->post('to_irspider', 'notice', $sender,
		sprintf("Currently tracking: %d bot%s, with %d pack%s",
			${\(&show_bots($sender, 2))},
			(${\(&show_bots($sender, 2))} > 1) ? 's' : '',
			${\(&show_packs())},
			(${\(&show_packs())} > 1) ? 's' : ''));
}

sub master_find
{
	my $input   = shift;
	my $sender  = shift;
	my $success = 0;
	my @buffer  = ();
	my $count   = 0;

    if (length($input) < $config{minlength})
	{
		$kernel->post('to_irspider', 'notice', $sender,
			"Error: Your search string must be at least $config{minlength} " .
			"characters in length.");
		mylog("search: returned error to $sender: search string too short.\n");
		return;
	}

	for my $b (sort (keys %bot))
	{
		my %i;
		
		for my $i (@{$bot{$b}{Pack}})
		{
			next if not $i->{Description};
			$count++;
			if ($i->{Description} =~ /$input/i)
			{
				$success++;
				push(@buffer, 
					sprintf("$success - Bot: %s, %s of %s slots open, " .
							"Pack: $count, Downloads: $i->{Downloads}, " .
							"Size: $i->{Size}",
						$b,
						($bot{$b}{SlotsCurrentlyAvail}),
						($bot{$b}{SlotsTotalAvailable})));
				push (@buffer, 
					"$success - Command: /msg $b xdcc send #$count");
				push(@buffer, 
					"$success - Description: $i->{Description}");
				push (@buffer, 
					"$success - Note: $i->{Note}") if $i->{Note};
				push(@buffer, ' ');
			}
		}
		$count = 0;
	} 
	if (not $success)
	{
		$kernel->post('to_irspider', 'notice', $sender,
			'Sorry, no matches were found.');
		mylog("search: returned no matches found to $sender.\n");
	}
	else
	{
		if ($success > $config{maxreturns})
		{
			$kernel->post('to_irspider', 'notice', $sender,
				"Error: Your query for '$input' resulted in too many matches.");
			$kernel->post('to_irspider', 'notice', $sender,
				'Either make your search more specific or request a masslist.');
			mylog("search: returned error to $sender: too many matches.\n");
			return;
		}
		$kernel->post('to_irspider', 'notice', $sender,
			"** Your search has returned $success results:");
		$kernel->post('to_irspider', 'notice', $sender, ' ');
		for (@buffer)
		{
			$kernel->post('to_irspider', 'notice', $sender, $_);
		}
		$kernel->post('to_irspider', 'notice', $sender,
			'** END of results');
		mylog("success: returned $success matches to $sender.\n");
	}
}

sub thetime
{
	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
		localtime(time);
	$mon++; $year=$year+1900;
	return sprintf "[%02d-%02d-%02d %2d:%02d:%02d] : ", $mday, $mon, $year, $hour, $min, $sec;
}

sub mylog
{
	print LOGFILE &thetime() . shift if $config{log};
}


sub verify_admin
{
	my $sender  = shift;
	my $type    = shift;
	my $success = 0;

	map {$_ =~ s/\*/\.\+/gi; $success++ if $sender =~ /$_/} @adminhosts;

	if ($success)
	{
		return 1;
	}
	else
	{
		$sender =~ s/(.+)!.+/$1/;
		$kernel->post('to_irspider', 'notice', $sender,
			"You must have admin privlidges to execute this command.") if
				$type == 1;
		mylog("admin: command from $sender failed: insufficient privs\n") if
			$type == 1;
		return 0;
	}
}

sub time_click
{
	my @buffer = ();

	$config{plist}-- if $config{plist} > 0;
	$config{respider}-- if $config{respider} > 0;

	if ($config{respider} == 0)
	{
		for my $b (sort (keys %bot))
		{
			push(@buffer, $b);
		}

		%bot = ();

		for (sort @buffer)
		{
		  $poe_kernel->post( 'to_irspider', 'privmsg', $_, 'xdcc list' );
		}

		$config{respider} = $respider_save;
	}

	if ($config{plist} == 0)
	{
		$poe_kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
			"** $config{nick} (irspider v$version) **  IRC Search Engine Smartbot");
		$poe_kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
			"** To request a search: \"/msg $config{nick} search <whatever>\"");
		$poe_kernel->post( 'to_irspider', 'privmsg', [$config{channel}],
			"** To request help: \"/msg $config{nick} help\"");
		$config{plist} = $plist_save;
	}
	alarm(60);
}

sub process_configfile
{
	my $type       = shift;
	my $old_nick   = '';
	my $old_server = '';

	$old_nick = $config{nick} if $type == 2;	
	$old_server = $server if $type == 2;

	open CONFIGFILE, '<./irspider.conf' or
		die "Cannot open config file for reading: $!\n";

	while (<CONFIGFILE>)
	{
		chomp;
		next if /^#/;    # skip comments
		next if /^\s*$/; # skip spaces
		if (/^server\s+(.+)$/)
		{
			push(@servers, $1);
			next;
		}
		elsif (/^adminhost\s+(.+)$/)
		{
			push(@adminhosts, $1);
			next;
		}
		my ($cmd, $args) = split;
		$config{$cmd} = $args;
	}

	close CONFIGFILE;

	if (not($config{log} and
			$config{nick} and
			$config{channel} and 
			$config{port} and
			$config{plist} and
			$config{respider} and 
			$config{maxreturns} and
			$config{minlength} and
			$config{antiflood} and
			$config{masslist_location} and
			$config{masslist_response} and
			@servers > 0))
	{
		die "Your configuration file is invalid. Restore it from the backup\n" .
			"copy, and try again.\n";
	}

	$kernel->post('to_irspider', 'nick', $config{nick}) if
		$type == 2 and $old_nick ne $config{nick};

	if ($type == 2 and $old_server ne $server)
	{
		mylog("rehash: server jumping..\n");
		$kernel->post( 'to_irspider', 'connect',
		{
			Debug    => 0,
			Nick     => $config{nick},
			Server   => $server,
			Port     => $config{port},
			Username => $config{nick},
			Ircname  => "/msg $config{nick} help",
		});
	}
}

# Change this '0' to '1' for lots of debugging information. debug => 0
POE::Component::IRC->new( 'to_irspider', debug => 0) or
	die "Can't instantiate new IRC component!\n";
POE::Session->new( 'main' => [qw(_start _stop irc_001 irc_kick irc_disconnected
                                 irc_error irc_msg irc_notice irc_353 irc_join
								 irc_nick irc_part irc_quit)]);

$SIG{'ALRM'} = \&time_click;
alarm(60);

$poe_kernel->run();

exit 0;
